import os, time, json
from bitcash import transaction, network
from gamechain.gc_types import GcPrivateKey
import cashaddress
from .. import utils, which_net
from . import gamechain_monitor

GCL_PREFIX = 1337
GC_PREFIX = 31337

OP_RETURN_PREFIX_BYTES = 0x6A
OP_RETURN_PREFIX = "6a"

OP_PUSHDATA = 0x4C
OP_PUSHDATA2 = 0x4D

FEE_AMOUNT = 250
MESSAGE_AMOUNT = 0
AMOUNT_UNITS = 'satoshi'
MESSAGE_LIMIT = 220  # 100

NEXT_TO_SPEND_TXID = "next_txid"


class GcCommClient:

    def __init__(self, private_key, save_file):
        self._private_key = private_key
        self._save_file = save_file

        if not os.path.exists(save_file):
            init_data = self._create_init_data()
            self._data = init_data
            self._save_state()

        self._load_state()

        self._gamechain_monitor = gamechain_monitor.GameChainMonitor(60)

    def _create_init_data(self):
        unspents = self._private_key.get_unspents()
        unspents.sort(key=lambda u: u.amount, reverse=True)
        next_to_use_tx = unspents[0]
        init_data = {
            NEXT_TO_SPEND_TXID: next_to_use_tx.txid
        }
        return init_data

    def _load_state(self):
        with open(self._save_file) as json_file:
            self._data = json.load(json_file)

    def _save_state(self):
        print("Saving state: %s" % self._data)
        with open(self._save_file, 'w') as outfile:
            json.dump(self._data, outfile)

    def send_message(self, to_addr, msg):
        tx_id = send_message(self._private_key, self._data[NEXT_TO_SPEND_TXID], to_addr, msg)

        self._data[NEXT_TO_SPEND_TXID] = tx_id
        self._save_state()

        return tx_id

    def stop(self):
        self._gamechain_monitor.stop()


def rx_join_op_returns(op_return_asms, lokad_prefix):
    def process_first_op_return(op_return_bytes):
        # pull out GC prefix
        if op_return_bytes[0] == OP_RETURN_PREFIX_BYTES and \
                op_return_bytes[1] == 0x04 and \
                op_return_bytes[2] == lokad_prefix[0] and \
                op_return_bytes[3] == lokad_prefix[1] and \
                op_return_bytes[4] == lokad_prefix[2] and \
                op_return_bytes[5] == lokad_prefix[3]:
            return op_return_bytes[6:]

        return None

    def process_subsequent_op_return(op_return_bytes):
        if op_return_bytes[0] == OP_RETURN_PREFIX_BYTES:
            return op_return_bytes[1:]

        return None

    def process_length_prefix(op_return_bytes):
        if op_return_bytes[0] == OP_PUSHDATA:
            size = op_return_bytes[1]
            return op_return_bytes[2:2 + size + 1]
        elif op_return_bytes[0] == OP_PUSHDATA2:
            size_bytes = op_return_bytes[1:3]
            size = int.from_bytes(size_bytes, byteorder='big')
            return op_return_bytes[3:3 + size + 1]

        return None

    op_return_asm_bytes = process_first_op_return(op_return_asms[0])
    if op_return_asm_bytes is None:
        return None

    if len(op_return_asms) > 1:
        for subsequent_op_return_asm_bytes in op_return_asms[1:]:
            processed_bytes = process_subsequent_op_return(subsequent_op_return_asm_bytes)
            if processed_bytes is None:
                return None
            op_return_asm_bytes += processed_bytes

    op_return_asm_bytes = process_length_prefix(op_return_asm_bytes)
    return op_return_asm_bytes


def get_fee_amount(op_return_data, msg_count=0):
    op_return_fee_multiplier = 1
    if msg_count > 1:
        op_return_fee_multiplier = 3

    return FEE_AMOUNT + (op_return_fee_multiplier * len(op_return_data))


def get_unspent_by_txid(sender_key, next_to_spend_txid):
    next_tx_found = False
    while not next_tx_found:
        unspents = sender_key.get_unspents()
        for unspent in unspents:
            if unspent.txid == next_to_spend_txid:
                return unspent

        if not next_tx_found:
            time.sleep(1)


def send_message(sender_key: GcPrivateKey, spend_from_txid, receiver_addr, op_ret_data):
    unspent = get_unspent_by_txid(sender_key, spend_from_txid)
    print(unspent)

    def chunk_data(data, size):
        return (data[i:i + size] for i in range(0, len(data), size))

    messages = []

    if op_ret_data:
        message_chunks = chunk_data(op_ret_data, MESSAGE_LIMIT)

        for message in message_chunks:
            messages.append((message, None))

    fee_amount = get_fee_amount(op_ret_data, len(messages))
    amount_back_to_sender = unspent.amount - fee_amount - MESSAGE_AMOUNT
    outputs = [
        (sender_key.address, amount_back_to_sender),
        (receiver_addr, MESSAGE_AMOUNT),
    ]

    outputs.extend(messages)

    tx_hex = transaction.create_p2pkh_transaction(sender_key.bitcash_key, [unspent], outputs)
    tx_id = transaction.calc_txid(tx_hex)

    if which_net.is_testnet():
        network.NetworkAPI.broadcast_tx_testnet(tx_hex)
    else:
        raise Exception("TX broadcasting only tested with testnet")

    return tx_id


def get_sender_receiver_op_returns_by_txid(txid):
    tx = utils.get_tx(txid)
    vouts = tx["vout"]
    addr_vouts = [vout for vout in vouts if "addresses" in vout["scriptPubKey"].keys()]
    sender_addr = which_net.ensure_prefixed_address_str(addr_vouts[0]["scriptPubKey"]["addresses"][0])
    receiver_addr = which_net.ensure_prefixed_address_str(addr_vouts[1]["scriptPubKey"]["addresses"][0])
    op_return_hexes = [vout["scriptPubKey"]["hex"] for vout in vouts if
                      vout["scriptPubKey"]["hex"].startswith(OP_RETURN_PREFIX)]

    op_return_asms_bytes = [bytes.fromhex(op_return_asm) for op_return_asm in op_return_hexes]
    return sender_addr, receiver_addr, op_return_asms_bytes


def check_if_message_is_for_addr(tx, addr):
    vouts = tx["vout"]
    if len(vouts) < 2:
        return False

    receiver_addr = next(iter(vouts[1].keys()))
    receiver_addr = cashaddress.convert.to_cash_address(receiver_addr)
    receiver_addr = which_net.ensure_prefixed_address_str(receiver_addr)

    return receiver_addr == addr




from ..comm import gc_comm
from . import gc_parser

MSG_STT = 0x01
MSG_TMT = 0x03
MSG_WIN = 0x04
MSG_LUZ = 0x05
MSG_DRW = 0x06

# OP_PUSHDATA = 0x4C
# OP_PUSHDATA2 = 0x4D

GC_LOKAD_PREFIX = [0x00, 0x03, 0x13, 0x37]


class GcMessage:
    def __init__(self, txid, sender_addr, receiver_addr, msg_type, msg, previous_txid):
        self.txid = txid
        self.sender_addr = sender_addr
        self.receiver_addr = receiver_addr
        self.msg_type = msg_type
        self.msg = msg
        self.previous_txid = previous_txid

    @property
    def txid_bytes(self):
        return bytes.fromhex(self.txid)


class MsgStt:
    def __init__(self, public_key_bytes, msg_data):
        self.public_key_bytes = public_key_bytes
        self.msg_data = msg_data


class MsgTmt:
    def __init__(self, msg_data):
        self.msg_data = msg_data

#
# OP_RETURN_PREFIX_BYTES = 0x6A
# OP_RETURN_PREFIX = "6a"


# def join_gc_op_returns(op_return_asms):
#     def process_first_op_return(op_return_bytes):
#         # pull out GC prefix
#         if op_return_bytes[0] == OP_RETURN_PREFIX_BYTES and \
#                 op_return_bytes[1] == 0x04 and \
#                 op_return_bytes[2] == 0x00 and \
#                 op_return_bytes[3] == 0x03 and \
#                 op_return_bytes[4] == 0x13 and \
#                 op_return_bytes[5] == 0x37:
#             return op_return_bytes[6:]
#
#         return None
#
#     def process_subsequent_op_return(op_return_bytes):
#         if op_return_bytes[0] == OP_RETURN_PREFIX_BYTES:
#             return op_return_bytes[1:]
#
#         return None
#
#     def process_length_prefix(op_return_bytes):
#         if op_return_bytes[0] == OP_PUSHDATA:
#             size = op_return_bytes[1]
#             return op_return_bytes[2:2 + size + 1]
#         elif op_return_bytes[0] == OP_PUSHDATA2:
#             size_bytes = op_return_bytes[1:3]
#             size = int.from_bytes(size_bytes, byteorder='big')
#             return op_return_bytes[3:3 + size + 1]
#
#         return None
#
#     op_return_asm_bytes = process_first_op_return(op_return_asms[0])
#     if op_return_asm_bytes is None:
#         return None
#
#     if len(op_return_asms) > 1:
#         for subsequent_op_return_asm_bytes in op_return_asms[1:]:
#             processed_bytes = process_subsequent_op_return(subsequent_op_return_asm_bytes)
#             if processed_bytes is None:
#                 return None
#             op_return_asm_bytes += processed_bytes
#
#     op_return_asm_bytes = process_length_prefix(op_return_asm_bytes)
#     return op_return_asm_bytes


def receive_message_by_txid(txid, sender_public_key=None) -> GcMessage:
    sender_addr, receiver_addr, op_return_asms = gc_comm.get_sender_receiver_op_returns_by_txid(txid)
    op_return_gc_asm_bytes = gc_comm.rx_join_op_returns(op_return_asms, GC_LOKAD_PREFIX)
    if op_return_gc_asm_bytes is None:
        print(f"NO GC MESSAGE TO RECEIVE IN TX {txid}")
        return None

    # print(f"TX_ID: {txid}")
    # print(f"MSG_ASM: {op_return_gc_asm_bytes}")

    msg_type, msg_contents, previous_txid = gc_parser.parse_gc_message(op_return_gc_asm_bytes, sender_public_key)

    return GcMessage(txid, sender_addr, receiver_addr, msg_type, msg_contents, previous_txid)


def receive_message_by_txid_with_pubkeys(txid, public_keys={}) -> GcMessage:
    sender_addr, receiver_addr, op_return_asms = gc_comm.get_sender_receiver_op_returns_by_txid(txid)
    op_return_gc_asm_bytes = gc_comm.rx_join_op_returns(op_return_asms, GC_LOKAD_PREFIX)
    if op_return_gc_asm_bytes is None:
        print(f"NO GC MESSAGE TO RECEIVE IN TX {txid}")
        return None

    # print(f"TX_ID: {txid}")
    # print(f"MSG_ASM: {op_return_gc_asm_bytes}")

    sender_public_key = public_keys[sender_addr]
    msg_type, msg_contents, previous_txid = gc_parser.parse_gc_message(op_return_gc_asm_bytes, sender_public_key)

    return GcMessage(txid, sender_addr, receiver_addr, msg_type, msg_contents, previous_txid)


def get_str_for_msg_type(msg_type):
    if msg_type == MSG_STT:
        return "STT"
    elif msg_type == MSG_TMT:
        return "TMT"
    elif msg_type == MSG_WIN:
        return "WIN"
    elif msg_type == MSG_LUZ:
        return "LUZ"
    elif msg_type == MSG_DRW:
        return "DRW"

    return "???"



import requests
import time
from . import which_net


def get_tx(txid) -> str:
    insight_url = which_net.get_bitpay_insight_api_url()
    url = f"{insight_url}/tx/{txid}"
    while True:
        response = requests.get(url)
        if response.status_code == 200:
            return response.json()
        else:
            time.sleep(0.3)


def get_transaction_ids_for_address(addr):
    insight_api_url = which_net.get_bitpay_insight_api_url()
    addr_url = f"{insight_api_url}/addr/" + which_net.ensure_prefixed_address_str(addr)
    addr_data = requests.get(addr_url).json()
    txids = addr_data["transactions"]
    return txids


def single(items, predicate, raiseExceptionOnNotFound=True):
    filtered = [i for i in items if predicate(i)]
    if len(filtered) == 0:
        if raiseExceptionOnNotFound:
            raise Exception("No item that matches predicate %s" % predicate)
        return None
    if len(filtered) > 1:
        raise Exception("More than one item matches predicate %s" % predicate)

    return filtered[0]


def where(items, predicate):
    filtered = [i for i in items if predicate(i)]
    return filtered


if __name__ == "__main__":
    tx = get_tx("91955d1d228453fac7113850eeab48bf31350f5e0767a7d2357ca47af1836cde")

    print(tx)



import json
from gamechain.comm import gc_comm
from gamechain.play import gc_message


def _get_prefix_bytes():
    return 0x0400031337.to_bytes(5, 'big')


def _get_op_pushdata_bytes(message_body_len):
    return gc_comm.OP_PUSHDATA2.to_bytes(1, 'big') + message_body_len.to_bytes(2, 'big')


def _get_version_bytes():
    return 0x01.to_bytes(1, 'big')


def pad_gameid_to32(game_id):
    padded = game_id
    while len(padded) < 32:
        padded += " "
    return padded


def create_init_game_str_bytes(game_id,
                               p1_addr, p1_public_key,
                               p2_addr, p2_public_key,
                               challenger_game_conditions) -> str:
    init_game_conditions = {
        "gameAddr": game_id,
        "players": [
            {
                "addr": p1_addr,
                "key": p1_public_key.hex()
            },
            {
                "addr": p2_addr,
                "key": p2_public_key.hex()
            }
        ],
        "setup": challenger_game_conditions
    }
    init_game_conditions_str_bytes = json.dumps(init_game_conditions).encode()
    return init_game_conditions_str_bytes


def create_set_the_table_message(initiator_key, game_id, init_msg_bytes):
    stt_msg_bytes = gc_message.MSG_STT.to_bytes(1, 'big')
    if len(game_id) != 32:
        raise Exception(f"game_id must be 32 bytes long '{game_id}'")

    game_id_bytes = game_id.encode()
    signed_game_id_bytes = initiator_key.sign(game_id_bytes)
    signed_game_id_size_bytes = len(signed_game_id_bytes).to_bytes(1, 'big')

    msg_size = len(init_msg_bytes).to_bytes(2, 'big')

    message_body = [
        stt_msg_bytes,
        initiator_key.public_key,
        game_id_bytes,
        signed_game_id_size_bytes,
        msg_size,
        signed_game_id_bytes,
        init_msg_bytes
    ]
    message_body_bytes = b''.join(message_body)
    message_body_len = len(message_body_bytes)

    message_parts = [
        _get_prefix_bytes(),
        _get_op_pushdata_bytes(message_body_len),
        _get_version_bytes(),
        message_body_bytes
    ]

    message = b''.join(message_parts)
    return message


def create_taking_my_turn_message(player_key, prev_txid, turn_msg_bytes):
    tmt_msgtype_bytes = gc_message.MSG_TMT.to_bytes(1, 'big')

    prev_txid_bytes = bytes.fromhex(prev_txid)
    signed_prev_txid_bytes = player_key.sign(prev_txid_bytes)
    signed_prev_txid_size_bytes = len(signed_prev_txid_bytes).to_bytes(1, 'big')

    msg_size = len(turn_msg_bytes).to_bytes(2, 'big')

    message_body = [
        tmt_msgtype_bytes,
        prev_txid_bytes,
        msg_size,
        signed_prev_txid_size_bytes,
        turn_msg_bytes,
        signed_prev_txid_bytes
    ]
    message_body_bytes = b''.join(message_body)
    message_body_len = len(message_body_bytes)

    message_parts = [
        _get_prefix_bytes(),
        _get_op_pushdata_bytes(message_body_len),
        _get_version_bytes(),
        message_body_bytes
    ]

    message = b''.join(message_parts)
    return message


def create_i_win_message(player_key, prev_txid, turn_msg_bytes):
    iwin_msgtype_bytes = gc_message.MSG_WIN.to_bytes(1, 'big')

    prev_txid_bytes = bytes.fromhex(prev_txid)
    signed_prev_txid_bytes = player_key.sign(prev_txid_bytes)
    signed_prev_txid_size_bytes = len(signed_prev_txid_bytes).to_bytes(1, 'big')

    msg_size = len(turn_msg_bytes).to_bytes(2, 'big')

    message_body = [
        iwin_msgtype_bytes,
        prev_txid_bytes,
        msg_size,
        signed_prev_txid_size_bytes,
        turn_msg_bytes,
        signed_prev_txid_bytes
    ]
    message_body_bytes = b''.join(message_body)
    message_body_len = len(message_body_bytes)

    message_parts = [
        _get_prefix_bytes(),
        _get_op_pushdata_bytes(message_body_len),
        _get_version_bytes(),
        message_body_bytes
    ]

    message = b''.join(message_parts)
    return message


def create_draw_message(player_key, prev_txid, turn_msg_bytes):
    draw_msgtype_bytes = gc_message.MSG_DRW.to_bytes(1, 'big')

    prev_txid_bytes = bytes.fromhex(prev_txid)
    signed_prev_txid_bytes = player_key.sign(prev_txid_bytes)
    signed_prev_txid_size_bytes = len(signed_prev_txid_bytes).to_bytes(1, 'big')

    msg_size = len(turn_msg_bytes).to_bytes(2, 'big')

    message_body = [
        draw_msgtype_bytes,
        prev_txid_bytes,
        msg_size,
        signed_prev_txid_size_bytes,
        turn_msg_bytes,
        signed_prev_txid_bytes
    ]
    message_body_bytes = b''.join(message_body)
    message_body_len = len(message_body_bytes)

    message_parts = [
        _get_prefix_bytes(),
        _get_op_pushdata_bytes(message_body_len),
        _get_version_bytes(),
        message_body_bytes
    ]

    message = b''.join(message_parts)
    return message



import time
from bitcash import transaction, network
import cashaddress
from .. import utils
from . import gcl_parser
from gamechain.comm import gc_comm
from gamechain import which_net

GCL_PREFIX = 1337
GC_PREFIX = 31337

MSG_LFG = 0x01
MSG_WTP = 0x02
MSG_ACC = 0x03
MSG_REJ = 0x04
MSG_CAN = 0x05

FEE_AMOUNT = 250
MESSAGE_AMOUNT = 0
AMOUNT_UNITS ='satoshi'
MESSAGE_LIMIT = 220  # 100

OP_PUSHDATA = 0x4C

GCL_LOKAD_PREFIX = [0x00, 0x00, 0x13, 0x37]


class GclMessage:
    def __init__(self, txid, sender_addr, receiver_addr, msg_type, msg):
        self.txid = txid
        self.sender_addr = which_net.ensure_prefixed_address_str(sender_addr)
        self.receiver_addr = which_net.ensure_prefixed_address_str(receiver_addr)
        self.msg_type = msg_type
        self.msg = msg

    @property
    def txid_bytes(self):
        return bytes.fromhex(self.txid)


class MsgLfg:
    def __init__(self, public_key_bytes, msg_data):
        self.public_key_bytes = public_key_bytes
        self.msg_data = msg_data


class MsgWtp:
    def __init__(self, public_key_bytes, lfg_txid_bytes, msg_data):
        self.public_key_bytes = public_key_bytes
        self.lfg_txid_bytes = lfg_txid_bytes
        self.msg_data = msg_data


class MsgAcc:
    def __init__(self, wtp_txid_bytes, msg_data):
        self.wtp_txid_bytes = wtp_txid_bytes
        self.msg_data = msg_data


class MsgCan:
    def __init__(self, wtp_txid_bytes, msg_data):
        self.wtp_txid_bytes = wtp_txid_bytes
        self.msg_data = msg_data


class MsgRej:
    def __init__(self, wtp_txid_bytes, msg_data):
        self.wtp_txid_bytes = wtp_txid_bytes
        self.msg_data = msg_data


def get_unspent_by_txid(sender_key, next_to_spend_txid):
    next_tx_found = False
    while not next_tx_found:
        unspents = sender_key.get_unspents()
        # unspents.sort(key=lambda x: x.amount, reverse=True)
        for unspent in unspents:
            if unspent.txid == next_to_spend_txid:
                return unspent

        if not next_tx_found:
            time.sleep(1)


def receive_message_by_txid(txid, sender_public_key=None) -> GclMessage:
    sender_addr, receiver_addr, op_return_asms = gc_comm.get_sender_receiver_op_returns_by_txid(txid)
    op_return_gcl_asm_bytes = gc_comm.rx_join_op_returns(op_return_asms, GCL_LOKAD_PREFIX)
    if op_return_gcl_asm_bytes is None:
        print(f"NO GC MESSAGE TO RECEIVE IN TX {txid}")
        return None

    # op_return_asm_bytes = op_return_asms[0]

    print(f"TX_ID: {txid}")
    print(f"MSG_ASM: {op_return_gcl_asm_bytes}")
    # msg_type, msg_contents = gcl_parser.parse_op_return(op_return_gc_asm_bytes, sender_public_key)
    msg_type, msg_contents = gcl_parser.parse_gcl_message(op_return_gcl_asm_bytes, sender_public_key)

    return GclMessage(txid, sender_addr, receiver_addr, msg_type, msg_contents)


def check_if_message_is_for_addr(tx, addr):
    vouts = tx["vout"]
    if len(vouts) < 2:
        return False

    receiver_addr = next(iter(vouts[1].keys()))
    receiver_addr = cashaddress.convert.to_cash_address(receiver_addr)
    receiver_addr = which_net.ensure_prefixed_address_str(receiver_addr)

    return receiver_addr == addr


if __name__ == "__main__":
    print("MSG")
    msg, sender, receiver = receive_message_by_txid("a90285a9270c980b860f6f0921ec887395778a400e1e3c850dc0fac8dabba23d")
    # h = int("0x" + msg, 16)
    print(msg)
    # h = bytes.fromhex(msg)
    # print(h)




#
# const transaction = new bch.Transaction()
#         .from(utxo)
#         .to(toAddr, 0)
#         .change(fromAddr)
#         .fee(FEE)
#         .addData(message)
#         .sign(fromPk);



from .. import utils
from gamechain.lobby import gamechain_lobby
from ..play import gc_engine, gc_message, gc_state
import json


def initialize_game(initiator_key, JSON_FILE, table_addr, game_id,
                    p1_addr, p1_public_key,
                    p2_addr, p2_public_key,
                    challenger_game_conditions) -> str:
    init_game_conditions = {
        "gameAddr": game_id,
        "players": [
            {
                "addr": p1_addr,
                "key": p1_public_key.hex()
            },
            {
                "addr": p2_addr,
                "key": p2_public_key.hex()
            }
        ],
        "setup": challenger_game_conditions
    }
    init_game_conditions_str_bytes = json.dumps(init_game_conditions).encode()

    gcl_client = gamechain_lobby.GameChainLobbyClient(initiator_key, JSON_FILE)
    txid = gcl_client.send_gc_message(table_addr, init_game_conditions_str_bytes)
    return f"bitcoincash-gc:{txid}"


def get_game_messages(table_addr, player_pubkeys):
    txids = utils.get_transaction_ids_for_address(table_addr)

    game_messages = []
    for txid in txids:
        msg = gc_message.receive_message_by_txid_with_pubkeys(txid, player_pubkeys)
        if msg is not None:
            game_messages.append(msg)

    return game_messages


def start_playing(init_game_msg, player_key, initiator_public_key, json_file, game_processor):
    # find game tx
    bch_init_txid = init_game_msg.split("bitcoincash-gc:")[1]

    # receive game STT message
    stt_message = gc_message.receive_message_by_txid(bch_init_txid, initiator_public_key)

    if stt_message is None or stt_message.msg_type != gc_message.MSG_STT:
        raise Exception(f"Unable to play game based on init_game_msg: {init_game_msg}")

    init_game_msg = stt_message.msg.msg_data

    game_processor.initialize_game(init_game_msg)
    player_pubkeys = game_processor.player_pubkeys
    table_addr = stt_message.receiver_addr

    game_messages = get_game_messages(table_addr, player_pubkeys)
    game_state_to_now = gc_state.build_gc_message_chain(game_messages)
    # game_turns = [msg for msg in game_state_to_now if msg.msg_type == gc_message.MSG_TMT]
    game_turns = [msg for msg in game_state_to_now if msg.msg_type != gc_message.MSG_STT]

    print("GM INIT >> " + game_state_to_now[0].msg.msg_data)
    for gt in game_turns:
        print("GM >> " + gt.msg.msg_data)

    engine = gc_engine.GameEngine(game_processor, player_key, player_pubkeys, table_addr, json_file)
    engine.initialize_game(stt_message, game_turns)
    engine.play()




_use_testnet: bool = True


def set_use_mainnet(use_it: bool):
    global _use_testnet
    _use_testnet = not use_it


def is_testnet() -> bool:
    return _use_testnet


def get_bitpay_insight_api_url() -> str:
    if _use_testnet:
        return f"https://test-bch-insight.bitpay.com/api"

    raise Exception("Insight API only tested/available with testnet")


def ensure_prefixed_address_str(addr: str) -> str:
    if is_testnet():
        if not addr.startswith("bchtest:"):
            return f"bchtest:{addr}"
    else:
        if not addr.startswith("bitcoincash:"):
            return f"bitcoincash:{addr}"

    return addr




from typing import Union
from bitcash import PrivateKey, PrivateKeyTestnet


class TxId:
    pass


class GcAddr:
    pass


class GcPrivateKey:

    def __init__(self, pk: Union[PrivateKey, PrivateKeyTestnet] = None):
        if pk is None:
            self._pk = PrivateKeyTestnet()
        else:
            self._pk = pk

    def get_unspents(self):
        return self._pk.get_unspents()

    @property
    def public_key(self):
        return self._pk.public_key

    @property
    def address(self):
        return self._pk.address

    def sign(self, data):
        return self._pk.sign(data)

    @property
    def bitcash_key(self):
        return self._pk



from .. import utils
from gamechain.play import gc_message


def build_gc_message_chain(messages):
    msg_chain = []
    stt_msg = utils.single(messages, lambda m: m.msg_type == gc_message.MSG_STT)
    msg_chain.append(stt_msg)

    curr_txid = stt_msg.txid
    next_msg = utils.single(messages, lambda m: m.previous_txid == curr_txid, False)

    while next_msg is not None:
        msg_chain.append(next_msg)

        curr_txid = next_msg.txid
        next_msg = utils.single(messages, lambda m: m.previous_txid == curr_txid, False)

    return msg_chain




import time
from gamechain.comm import gc_comm
from gamechain.play import gc_message, gc_message_builder, gc_state, gamechain


class GameEngine:

    def __init__(self, game_processor, player_key, player_pubkeys, table_addr, json_file):
        self.game_processor = game_processor
        self.player_key = player_key
        self.player_pubkeys = player_pubkeys
        self.table_addr = table_addr
        self.json_file = json_file
        self._processed_txids = []

    def initialize_game(self, stt_msg, game_turns):
        self.game_processor.initialize_game(stt_msg.msg.msg_data)
        self._processed_txids.append(stt_msg.txid)
        for gt in game_turns:
            turn_msg = gt.msg.msg_data
            player_addr = gt.sender_addr
            self.game_processor.process_player_turn(player_addr, gt.msg_type, turn_msg)
            self._processed_txids.append(gt.txid)

    def play(self):
        gcc = gc_comm.GcCommClient(self.player_key, self.json_file)
        while not self.game_processor.is_game_complete():
            if self.game_processor.is_player_turn(self.player_key.address):
                prev_turn_txid = self._processed_txids[-1]
                turn_msg_type, turn_msg = self.game_processor.wait_for_player_move(self.player_key.address)
                turn_msg_bytes = turn_msg.encode()
                if turn_msg_type == gc_message.MSG_TMT:
                    to_send = gc_message_builder.create_taking_my_turn_message(self.player_key, prev_turn_txid, turn_msg_bytes)
                elif turn_msg_type == gc_message.MSG_WIN:
                    to_send = gc_message_builder.create_i_win_message(self.player_key, prev_turn_txid, turn_msg_bytes)
                elif turn_msg_type == gc_message.MSG_DRW:
                    to_send = gc_message_builder.create_draw_message(self.player_key, prev_turn_txid, turn_msg_bytes)
                else:
                    raise Exception("Invalid turn_msg_type: %s" % turn_msg_type)
                turn_txid = gcc.send_message(self.table_addr, to_send)
                self._processed_txids.append(turn_txid)
            else:
                # check for and process new messages
                print("Checking for new game messages")
                game_messages = gamechain.get_game_messages(self.table_addr, self.player_pubkeys)
                gc_message_chain = gc_state.build_gc_message_chain(game_messages)
                new_messages = [msg for msg in gc_message_chain if msg.txid not in self._processed_txids]
                if len(new_messages) == 0:
                    time.sleep(2)
                else:
                    for next_msg in new_messages:
                        self.game_processor.process_player_turn(next_msg.sender_addr, next_msg.msg_type, next_msg.msg.msg_data)
                        self._processed_txids.append(next_msg.txid)

        gcc.stop()
        print("Game over! Winner is %s" % self.game_processor.game_winner)



import json
from itertools import cycle
from gamechain.play import gc_message


class TicTacToeProcessor:

    def initialize_game(self, init_game_msg):
        game_data = json.loads(init_game_msg)
        all_player_data = game_data["players"]
        players = [(pd["addr"], bytes.fromhex(pd["key"])) for pd in all_player_data]
        self.players = players
        self.player_cycle = cycle(players)
        self.next_to_act_addr = self.get_next_to_act()

        self.game_state = "---|---|---"
        self.x_addr = players[0][0]
        self.o_addr = players[1][0]

        self.game_over = False
        self.game_winner = None

    @property
    def player_pubkeys(self):
        return dict(self.players)

    @property
    def player_order(self):
        return [p[0] for p in self.players]

    def process_player_turn(self, player_addr, msg_type, turn_msg):
        if player_addr != self.next_to_act_addr:
            raise Exception("Invalid next to act addr: " + player_addr)

        if not self.check_valid(turn_msg):
            raise Exception("Invalid move! " + turn_msg)

        if msg_type == gc_message.MSG_TMT:
            self.game_state = turn_msg
            self.next_to_act_addr = self.get_next_to_act()

        if msg_type == gc_message.MSG_DRW:
            if self._is_game_complete(turn_msg):
                self.report_to_player("Game is a draw")
                self.game_over = True
            else:
                self.report_to_player("Draw declared but game is not complete!")

        if msg_type == gc_message.MSG_WIN:
            check_to_win_symbol = self.get_player_symbol(player_addr)
            if self.is_symbol_winner(turn_msg, check_to_win_symbol):
                self.report_to_player("Winner is %s" % check_to_win_symbol)
                self.game_winner = player_addr
                self.game_over = True
            else:
                self.report_to_player("Win declared by %s but game is not complete!" % check_to_win_symbol)

    def check_valid(self, turn_msg):
        if turn_msg is None:
            return False
        if len(turn_msg) != len("---|---|---"):
            return False
        if turn_msg[3] != "|" or turn_msg[7] != "|":
            return False
        return True

    def get_player_symbol(self, player_addr):
        player_symbol = "?"
        if player_addr == self.x_addr:
            player_symbol = "X"
        if player_addr == self.o_addr:
            player_symbol = "O"
        return player_symbol

    def create_input_message(self, game_state, player_addr):
        player_symbol = self.get_player_symbol(player_addr)
        msg_strs = ["\n", "You are " + player_symbol, "\n"]

        rows = game_state.split("|")
        for r in rows:
            msg_strs.append(r)
            msg_strs.append("\n")

        msg_strs.append("Choose grid location for " + player_symbol + " >> ")
        return "".join(msg_strs)

    def wait_for_player_move(self, player_addr):
        valid = False
        while not valid:
            input_message = self.create_input_message(self.game_state, player_addr)
            next_move = input(input_message)
            updated_game_state = self.process_move(next_move, player_addr)
            valid = self.check_valid(updated_game_state)

        self.game_state = updated_game_state
        self.next_to_act_addr = self.get_next_to_act()
        player_gc_message_type = self.get_player_gc_message_type(updated_game_state, player_addr)

        if player_gc_message_type == gc_message.MSG_WIN:
            self.game_over = True
            self.game_winner = player_addr

        return player_gc_message_type, updated_game_state

    def process_move(self, move, player_addr):
        player_symbol = self.get_player_symbol(player_addr)
        updated_game_state = self.game_state
        move_index = self.get_index_for_move(move)
        if move_index is None:
            return None
        if updated_game_state[move_index] != "-":
            return None
        updated_game_state = updated_game_state[:move_index] + player_symbol + updated_game_state[move_index+1:]
        return updated_game_state

    def get_index_for_move(self, move):
        if len(move) != 2:
            return None

        if move.startswith("A"):
            col_move_index = 0
        elif move.startswith("B"):
            col_move_index = 1
        elif move.startswith("C"):
            col_move_index = 2
        else:
            return None

        if move.endswith("1"):
            row_move_index = 0
        elif move.endswith("2"):
            row_move_index = 1
        elif move.endswith("3"):
            row_move_index = 2
        else:
            return None

        total_index = (4 * row_move_index) + col_move_index
        return total_index

    def report_to_player(self, info):
        input("REPORT >> " + info)

    def _is_game_complete(self, game_state):
        any_moves_left = "-" in game_state
        if not any_moves_left:
            return True
        return False

    def is_game_complete(self):
        return self._is_game_complete(self.game_state) or self.game_over

    def is_symbol_winner(self, game_state, symbol):
        if self._is_top_row_winner(game_state, symbol) or \
            self._is_middle_row_winner(game_state, symbol) or \
            self._is_bottom_row_winner(game_state, symbol) or \
            self._is_left_column_winner(game_state, symbol) or \
            self._is_center_column_winner(game_state, symbol) or \
            self._is_right_column_winner(game_state, symbol) or \
            self._is_top_left_diagonal_winner(game_state, symbol) or \
            self._is_top_right_diagonal_winner(game_state, symbol):
                return True
        return False

    def get_player_gc_message_type(self, game_state, player_addr):
        player_symbol = self.get_player_symbol(player_addr)
        if self.is_symbol_winner(game_state, player_symbol):
            return gc_message.MSG_WIN

        if self._is_game_complete(game_state):
            return gc_message.MSG_DRW

        return gc_message.MSG_TMT

    def _get_row(self, row_index, game_state):
        return game_state.split("|")[row_index]

    def _get_column(self, col_index, game_state):
        row1 = self._get_row(0, game_state)
        row2 = self._get_row(1, game_state)
        row3 = self._get_row(2, game_state)

        col = row1[col_index] + row2[col_index] + row3[col_index]
        return col

    def _get_top_left_diagonal(self, game_state):
        top_left = self._get_row(0, game_state)[0]
        middle = self._get_row(1, game_state)[1]
        bottom_right = self._get_row(2, game_state)[2]

        diagonal = top_left + middle + bottom_right
        return diagonal

    def _is_top_left_diagonal_winner(self, game_state, symbol):
        diag_sequence = self._get_top_left_diagonal(game_state)
        symbol_sequence = symbol * 3
        if diag_sequence == symbol_sequence:
            return True
        return False

    def _is_top_right_diagonal_winner(self, game_state, symbol):
        diag_sequence = self._get_top_right_diagonal(game_state)
        symbol_sequence = symbol * 3
        if diag_sequence == symbol_sequence:
            return True
        return False

    def _get_top_right_diagonal(self, game_state):
        top_right = self._get_row(0, game_state)[2]
        middle = self._get_row(1, game_state)[1]
        bottom_left = self._get_row(2, game_state)[0]

        diagonal = bottom_left + middle + top_right
        return diagonal

    def _check_row_winner(self, game_state, symbol, row_index):
        row_sequence = self._get_row(row_index, game_state)
        symbol_sequence = symbol * 3
        if row_sequence == symbol_sequence:
            return True
        return False

    def _is_top_row_winner(self, game_state, symbol):
        return self._check_row_winner(game_state, symbol, 0)

    def _is_middle_row_winner(self, game_state, symbol):
        return self._check_row_winner(game_state, symbol, 1)

    def _is_bottom_row_winner(self, game_state, symbol):
        return self._check_row_winner(game_state, symbol, 2)

    def _check_column_winner(self, game_state, symbol, col_index):
        col_sequence = self._get_column(col_index, game_state)
        symbol_sequence = symbol * 3
        if col_sequence == symbol_sequence:
            return True
        return False

    def _is_left_column_winner(self, game_state, symbol):
        return self._check_column_winner(game_state, symbol, 0)

    def _is_center_column_winner(self, game_state, symbol):
        return self._check_column_winner(game_state, symbol, 1)

    def _is_right_column_winner(self, game_state, symbol):
        return self._check_column_winner(game_state, symbol, 2)

    def is_player_turn(self, player_addr):
        return self.next_to_act_addr == player_addr

    def get_next_to_act(self):
        return next(self.player_cycle)[0]




from bitcash import format
from . import gc_message


def parse_stt(msg_payload):
    initiator_key_bytes = msg_payload[:33]
    game_id_bytes = msg_payload[33:33+32]

    sizes_offset = 33 + 32
    signed_game_id_size = msg_payload[sizes_offset]
    msg_bytes_size_bytes = msg_payload[sizes_offset + 1:sizes_offset + 1 + 2]
    msg_size = int.from_bytes(msg_bytes_size_bytes, byteorder='big')

    messages_offset = sizes_offset + 3
    signed_game_id = msg_payload[messages_offset:messages_offset + signed_game_id_size]
    msg_data = msg_payload[messages_offset + signed_game_id_size: messages_offset + signed_game_id_size + msg_size]

    valid = format.verify_sig(signed_game_id, game_id_bytes, initiator_key_bytes)
    if valid:
        msg_data_str = msg_data.decode()
        msg_stt = gc_message.MsgStt(initiator_key_bytes, msg_data_str)
        return gc_message.MSG_STT, msg_stt, None

    return None


def parse_tmt(msg_payload, sender_public_key):
    previous_txid_bytes = msg_payload[:32]

    sizes_offset = 32
    msg_bytes_size_bytes = msg_payload[sizes_offset:sizes_offset + 2]
    msg_size = int.from_bytes(msg_bytes_size_bytes, byteorder='big')
    signed_previous_txid_size = msg_payload[sizes_offset + 2]

    after_sizes_offset = 32 + 3
    game_turn_ops = msg_payload[after_sizes_offset:after_sizes_offset + msg_size]

    after_game_turn_ops_offset = after_sizes_offset + msg_size
    signed_previous_txid = msg_payload[after_game_turn_ops_offset:after_game_turn_ops_offset + signed_previous_txid_size + 1]

    valid = format.verify_sig(signed_previous_txid, previous_txid_bytes, sender_public_key)
    if valid:
        game_turn_ops_str = game_turn_ops.decode()
        msg_tmt = gc_message.MsgTmt(game_turn_ops_str)
        previous_txid = previous_txid_bytes.hex()
        return gc_message.MSG_TMT, msg_tmt, previous_txid

    return None


def parse_win(msg_payload, sender_public_key):
    previous_txid_bytes = msg_payload[:32]

    sizes_offset = 32
    msg_bytes_size_bytes = msg_payload[sizes_offset:sizes_offset + 2]
    msg_size = int.from_bytes(msg_bytes_size_bytes, byteorder='big')
    signed_previous_txid_size = msg_payload[sizes_offset + 2]

    after_sizes_offset = 32 + 3
    game_turn_ops = msg_payload[after_sizes_offset:after_sizes_offset + msg_size]

    after_game_turn_ops_offset = after_sizes_offset + msg_size
    signed_previous_txid = msg_payload[after_game_turn_ops_offset:after_game_turn_ops_offset + signed_previous_txid_size + 1]

    valid = format.verify_sig(signed_previous_txid, previous_txid_bytes, sender_public_key)
    if valid:
        game_turn_ops_str = game_turn_ops.decode()
        msg_win = gc_message.MsgTmt(game_turn_ops_str)
        previous_txid = previous_txid_bytes.hex()
        return gc_message.MSG_WIN, msg_win, previous_txid

    return None


def parse_draw(msg_payload, sender_public_key):
    previous_txid_bytes = msg_payload[:32]

    sizes_offset = 32
    msg_bytes_size_bytes = msg_payload[sizes_offset:sizes_offset + 2]
    msg_size = int.from_bytes(msg_bytes_size_bytes, byteorder='big')
    signed_previous_txid_size = msg_payload[sizes_offset + 2]

    after_sizes_offset = 32 + 3
    game_turn_ops = msg_payload[after_sizes_offset:after_sizes_offset + msg_size]

    after_game_turn_ops_offset = after_sizes_offset + msg_size
    signed_previous_txid = msg_payload[after_game_turn_ops_offset:after_game_turn_ops_offset + signed_previous_txid_size + 1]

    valid = format.verify_sig(signed_previous_txid, previous_txid_bytes, sender_public_key)
    if valid:
        game_turn_ops_str = game_turn_ops.decode()
        msg_draw = gc_message.MsgTmt(game_turn_ops_str)
        previous_txid = previous_txid_bytes.hex()
        return gc_message.MSG_DRW, msg_draw, previous_txid

    return None


def parse_gc_message(gc_bytes, sender_public_key):
    if gc_bytes[0] != 0x01:
        return None

    msg_type = gc_bytes[1]
    msg_payload = gc_bytes[2:]

    if msg_type == gc_message.MSG_STT:
        return parse_stt(msg_payload)

    if msg_type == gc_message.MSG_TMT:
        return parse_tmt(msg_payload, sender_public_key)

    if msg_type == gc_message.MSG_WIN:
        return parse_win(msg_payload, sender_public_key)

    if msg_type == gc_message.MSG_DRW:
        return parse_draw(msg_payload, sender_public_key)

    return None
